// the heart object

#include "heart.h"
#include "level.h"
#include "arrow.h"
#include "rocket.h"
#include "enemy.h"
#include "particle.h"
#include "sound.h"
#include "gameplay.h"



// global pointer to the heart
Heart *heart = NULL;



Heart::Heart(double x, double y)
    : Thing(), mFront(1,0)
{
    mFront.set(0,-1);
    mPos.set(x,y);
    mVel.set(0,0);
    mRadius = 20;
    mAngVel = 0;
	mFireCounter = 0;
	mScore = 0;
}



// heart update function
bool Heart::update()
{
    if(!game_over)
        mAngVel *= 0.95;
    else
        mAngVel *= 0.999;

    if(!game_over) {
        if(key[KEY_LEFT]) {
            if(mAngVel <= 0.001)
                mAngVel = MAX(mAngVel-0.003, -0.1);
            else
                mAngVel *= 0.6;
        }
        else if(key[KEY_RIGHT]) {
            if(mAngVel >= -0.001)
                mAngVel = MIN(mAngVel+0.003, 0.1);
            else
                mAngVel *= 0.6;
        }
    }

    mFront.set( mFront.x * cos(mAngVel) - mFront.y * sin(mAngVel),
                mFront.y * cos(mAngVel) + mFront.x * sin(mAngVel));
    mFront.normalize();

    if(!game_over) {
        if(key[KEY_UP]) {
            mVel += mFront * 0.1;
            if (use_particles) {
                Vector right (mFront.y, -mFront.x);
                for (int i = 0; i < 3; i++) {
                    Vector pos = mPos - mFront * mRadius + right * ((rand() & 0xff) - 0x80) / 127.0 * mRadius * 0.5;
                    Vector vel = mFront * -3;
                    vel.x += 2 * ((rand() & 0xff) - 0x80) / 127.0;
                    vel.y += 2 * ((rand() & 0xff) - 0x80) / 127.0;
                    particles.add (Particle (pos, vel, 10, 0.01, Particle::TYPE_THRUST, true));
                }
            }
        } 
        
        if(key[KEY_DOWN]) {
            mVel -= mFront * 0.025;
            if (use_particles) {
                Vector right (mFront.y, -mFront.x);
                for (int i = 0; i < 3; i++) {
					Vector pos = mPos + mFront * mRadius + right * ((rand() & 0xff) - 0x80) / 127.0 * mRadius * 0.25;
					Vector vel = mFront;
                    vel.x += ((rand() & 0xff) - 0x80) / 127.0;
                    vel.y += ((rand() & 0xff) - 0x80) / 127.0;
                    particles.add (Particle (pos, vel, 10, 0.01, Particle::TYPE_THRUST, true));
                }
            }
        }
        
		update_engine((key[KEY_UP] != 0), (key[KEY_DOWN] != 0));

        if (mFireCounter) mFireCounter--;
        if ((key[KEY_RCONTROL] || key[KEY_LCONTROL] || key[KEY_SPACE] || key[KEY_ENTER]) && !mFireCounter) {
            mFireCounter = 40;
            Vector pos = mPos+mFront*10;
            Vector vel = mVel+mFront*5;
            arrows.add (Arrow (pos.x, pos.y, vel.x, vel.y));
            trigger_particles(TRIGGER_FIRE_ARROW, pos, vel);
        }
    }
	else
		update_engine(false, false);

    // gravity
    mVel += Vector(0, 0.01);

    // friction
    mVel = mVel * 0.99;

    mPos += mVel;


    // Check world collisions

    vector<LevelCollision> collisions;

    if(check_level_collision(this, collisions)) {
        Vector extract(0,0);
        bool d=false;

        vector<LevelCollision>::iterator it = collisions.begin();
        while(it != collisions.end()) {
            LevelCollision& col = *it;

            float amount = col.normal | mVel;

            if(amount < 0) {
                mVel -= (col.normal * amount) * 1.6;

                Vector newextract = col.normal * (mRadius - col.dist);

                extract = extract * 0.5 + newextract * 0.5;
		
				trigger_particles(TRIGGER_PLAYER_WALL, mPos - col.normal * col.dist, col.normal * col.dist);
            }

            it++;
        }

        mPos += extract;
    }


    // Check rocket collisions
    ThingList<Rocket>::iterator it;
    
    for(it = rockets.begin(); it != rockets.end(); it++) {
        Rocket& rocket = *it;

        // trivial rejects first
        if(rocket.mPos.x + rocket.mRadius < mPos.x - mRadius)
            continue;

        if(rocket.mPos.x - rocket.mRadius > mPos.x + mRadius)
            continue;

        float distsq = (rocket.mPos - mPos).square_magnitude();

        if(distsq < (rocket.mRadius + mRadius) * (rocket.mRadius + mRadius) ) {
            Vector n = ~(rocket.mPos - mPos);

            Vector relvel = rocket.mVel - mVel;

            float a = relvel | n;

            mVel += n * a;
            rocket.mVel -= n * a * 2;

            float penamount = sqrt( (rocket.mRadius + mRadius) * (rocket.mRadius + mRadius) - distsq);
            mPos -= n * penamount * 0.5;
            rocket.mPos += n * penamount * 0.5;

            rocket.mOwnerID = -1;

			trigger_particles(TRIGGER_PLAYER_ROCKET, (mPos + rocket.mPos) / 2, n*a);
        }
    }


    if(!game_over) {
        // Gain score
        mScore += enemies.size();
    }


	return true;
}



// heart render function
void Heart::render(BITMAP *bmp)
{
    Vector right(-mFront.y, mFront.x);

    Vector p[3];
    
    if(!game_over) {
        if (key[KEY_UP]) {
            p[0] = mPos + (mFront * 10) + scroll;
            p[1] = mPos - (mFront * 25) + (right * 12) + scroll;
            p[2] = mPos - (mFront * 25) - (right * 12) + scroll;
            
            triangle(bmp, p[0].x, p[0].y, p[1].x, p[1].y, p[2].x, p[2].y, makecol(255, 128, 0));
        }
        
        if (key[KEY_DOWN]) {
            p[0] = mPos + (mFront * 10) + scroll;
            p[1] = mPos + (mFront * 22) + (right * 12) + scroll;
            p[2] = mPos + (mFront * 22) - (right * 12) + scroll;
            
            triangle(bmp, p[0].x, p[0].y, p[1].x, p[1].y, p[2].x, p[2].y, makecol(255, 128, 0));
        }
    }

    circlefill(bmp, mPos.x + scroll.x, mPos.y + scroll.y, mRadius, makecol(0, 48, 192));

    p[0] = mPos + (mFront * 30) + scroll;
    p[1] = mPos - (mFront * 20) + (right * 20) + scroll;
    p[2] = mPos - (mFront * 20) - (right * 20) + scroll;

    triangle(bmp, p[0].x, p[0].y, p[1].x, p[1].y, p[2].x, p[2].y, makecol(0, 64, 255));

    p[0] = mPos + (mFront * 10) + scroll;
    p[1] = mPos - (mFront * 20) + (right * 10) + scroll;
    p[2] = mPos - (mFront * 20) - (right * 10) + scroll;

    triangle(bmp, p[0].x, p[0].y, p[1].x, p[1].y, p[2].x, p[2].y, makecol(0, 48, 192));
}

