// level (environment) code

#include "level.h"
#include "heart.h"
#include "lighting.h"
#include "plasma.h"
#include "enemy.h"
#include "panel.h"



// levels are defined in external .cpp files
extern Level level1;
extern Level level2;
extern Level level3;
extern Level level4;
extern Level level5;
extern Level level6;

// list of all levels
Level *level_table[] =
{
	&level1,
	&level2,
	&level3,
	&level4,
    &level5,
    &level6,
	NULL
};



// the current level
Level *level = NULL;



// scrolling position
Vector scroll;
Vector scroll_vel;



// a polygon edge (start, end, and horizontal slope)
class Edge
{
public:
	
	// construktor
	Edge(int *data)
	{
		x1 = *(data++);
		y1 = *(data++);

		x2 = *(data++);
		y2 = *(data++);

        normal.set(y2-y1, x1-x2);
        normal.normalize();

		if (y1 > y2) {
			swap(x1, x2);
			swap(y1, y2);
		}
		
		dx = (float)(x2 - x1) / (float)(y2 - y1);
	}


	// sort by top Y
	bool operator < (const Edge &other) const
	{
		return (y1 < other.y1);
	}


	// start point
	int x1, y1;
	
	// end point
	int x2, y2;

	// horizontal gradient
	float dx;

    // err...the normal
    Vector normal;
};



// list of edges, sorted by their top Y coordinate
static vector<Edge> edges;



// sets up the list of edge segments
static void init_edges()
{
	edges.clear();
	
	for (int i=0; i<level->mNumEdges; i++)
		edges.push_back(Edge(level->mPoints+i*4));

	sort(edges.begin(), edges.end());
}



// active edge structure
class ActiveEdge
{
public:
	
	// construktor
	ActiveEdge(const Edge &edge)
	{
		x = edge.x1;
		
		dx = edge.dx;

		end_y = edge.y2;
	}
	

	// sort by X
	bool operator < (const ActiveEdge &other) const
	{
		return (x < other.x);
	}


	// data
	float x, dx;

	int end_y;
};



// rasterises the polygon environments
class Rasteriser
{
public:
	
	// construktor
	Rasteriser(int y)
	{
		mNextEdge = edges.begin();
		
		mY = MIN(y, mNextEdge->y1) - 1;
		
		mStartY = y;
	}

	
	
	// advance to the next Y location
	bool get_line()
	{
		mY++;
		
		// update active edges, and retire old ones
		list<ActiveEdge>::iterator it = mActive.begin();
		
		while (it != mActive.end()) {
			if (mY < it->end_y) {
				it->x += it->dx;
				it++;
			}
			else
				it = mActive.erase(it);
		}

		// add newly active edges
		while ((mNextEdge != edges.end()) && (mNextEdge->y1 <= mY)) {
			if ((mNextEdge->y1 != mNextEdge->y2) && (mNextEdge->y2 >= mStartY))
				mActive.push_back(ActiveEdge(*mNextEdge));

			mNextEdge++;
		}
		
		if (mY < mStartY)
			return true;
		
		// sort the active edge list
		mActive.sort();
		
		mNextActive = mActive.begin();
		
		mFlipflop = -1;
		
		// is there anything left to do?
		return ((!mActive.empty()) || (mNextEdge != edges.end()));
	}



	// look up the next horizontal segment
	bool get_segment(int &x1, int &x2, bool &inside)
	{
		if (mFlipflop < 0) {
			// the first segment starts at -inf
			x1 = -65536;
		}
		else {
			// otherwise we continue from a previous segment
			if (mNextActive == mActive.end())
				return false;

			x1 = (int)mNextActive->x;

			mNextActive++;
		}

		if (mNextActive == mActive.end()) {
			// the last segment continues to +inf
			x2 = 65536;
		}
		else {
			// otherwise we just go up to the start of the next segment
			x2 = (int)mNextActive->x - 1;
		}

		// toggle the inside/outside flag
		inside = mFlipflop = !mFlipflop;

		return true;
	}



	// first Y position we are interested in
	int mStartY;


	// current Y position
	int mY;


	// next inactive edge
	vector<Edge>::iterator mNextEdge;
	
	
	// list of active edges
	list<ActiveEdge> mActive;


	// current edge segment
	list<ActiveEdge>::iterator mNextActive;


	// inside, outside?
	int mFlipflop;
};



// init the current level
void init_level()
{
	init_edges();

	scroll.set(0, 0);
	scroll_vel.set(0, 0);

	for (int i = 0; i < level->mNumObjects; i++)
		switch (level->mObjData[i].type) {

			case OBJTYPE_SHIP:
				{
					ASSERT (!heart);
					heart = new Heart (level->mObjData[i].x, level->mObjData[i].y);
					scroll.x = -level->mObjData[i].x + SCREEN_W/2;
					scroll.y = -level->mObjData[i].y + SCREEN_H/2;
				}
				break;

			case OBJTYPE_ENEMY:
				{
					Enemy enemy(level->mObjData[i].x, level->mObjData[i].y, level->mObjData[i].data[0], level->mObjData[i].id, level->mObjData[i].targets, (1+level->mObjData[i].data[1]) * 0.00001);
					player_panel.push_back(PlayerPanel(enemy.mName));
					enemy.mPanel = &player_panel.back();
					player_panel.sort();
    	            enemies.add(enemy);
				}
				break;

			default:
				TRACE ("object %d of unknown type %d\n", i, level->mObjData[i].type);
		}
}



// update the level (scrolling, etc)
void update_level()
{
    Vector center;

    float vamount = (heart->mVel | Vector(0,1));
    float hamount = (heart->mVel | Vector(1,0));

    center.x = 320 - (hamount * 100);
    center.y = 240 - (vamount * 100);

    Vector wanted_vel = ((center - heart->mPos) - scroll) * 0.01;

    scroll_vel = wanted_vel * 0.1 + scroll_vel * 0.9;

    scroll += scroll_vel;
}



// draw the current level
void render_level(BITMAP *bmp)
{
	render_lights();

	int scroll_x = (int)scroll.x;
	int scroll_y = (int)scroll.y;

	int x1, x2;

	bool inside;
	
	Rasteriser rast(-scroll_y);

	while (rast.mY+scroll_y < SCREEN_H-1) {
		rast.get_line();

		int y = rast.mY + scroll_y;

		if (y < 0)
			continue;

		while (rast.get_segment(x1, x2, inside)) {
			x1 += scroll_x;
			x2 += scroll_x;

			if (x1 < 0)
				x1 = 0;
			
			if (x2 >= SCREEN_W)
				x2 = SCREEN_W-1;

			if (x2 >= x1) {
				if (inside)
					hline(bmp, x1, y, x2, 0);
				else
					fill_scanline(bmp, x1, y, x2-x1, x1-scroll_x, y-scroll_y);
			}
		}
	}
}



bool check_level_collision(Thing* thing, vector<LevelCollision>& collisions)
{
    collisions.clear();

    vector<Edge>::iterator it;


    for(it = edges.begin(); it != edges.end(); it++) {
        Edge& edge = *it;


        // Early giveups (do y first, since edges are sorted by y)

        if(edge.y1 < thing->mPos.y - thing->mRadius && 
           edge.y2 < thing->mPos.y - thing->mRadius)
            continue;

        if(edge.y1 > thing->mPos.y + thing->mRadius &&
           edge.y2 > thing->mPos.y + thing->mRadius)
            continue;

        if(edge.x1 < thing->mPos.x - thing->mRadius && 
           edge.x2 < thing->mPos.x - thing->mRadius)
            continue;

        if(edge.x1 > thing->mPos.x + thing->mRadius &&
           edge.x2 > thing->mPos.x + thing->mRadius)
            continue;

        
        Vector normal = edge.normal;


        float dist = (thing->mPos - Vector(edge.x1, edge.y1)) | normal;
        
        if(dist < 0)
            continue;

        if(dist < thing->mRadius) {

            Vector linedir(-normal.y, normal.x);
            Vector p1(edge.x1, edge.y1);
            Vector p2(edge.x2, edge.y2); 

            float t = (thing->mPos - p1) | linedir;

            float linelength = (p2-p1) | linedir;

            if(linelength < 0) {
                linelength = -linelength;
                t = -t;
            }

            if(t < -thing->mRadius || t > linelength + thing->mRadius) {
                continue;
            }
            else if(t < 0) {
                dist = (p1 - thing->mPos).magnitude();
                normal = ~(-p1 + thing->mPos);

                if(dist >= thing->mRadius)
					continue;
            }
            else if(t > linelength) {
                dist = (p2 - thing->mPos).magnitude();
                normal = ~(-p2 + thing->mPos);

                if(dist >= thing->mRadius)
                	continue;
            }

            collisions.push_back(LevelCollision(dist,normal));
        }
        
    }

    if(collisions.empty())
        return false;
    else
        return true;
}


bool check_level_collision_line(const Vector &p1, const Vector &p2, vector<LevelCollision>& collisions, float tolerance)
{
    collisions.clear();

    vector<Edge>::iterator it;

	float dist = -1; // distance until the collision, used to find the nearest collision only
	Edge *coll_edge = NULL; // nearest colliding edge
	float depth = 0;


    for(it = edges.begin(); it != edges.end(); it++) {
        Edge& edge = *it;


        // Early giveups (do y first, since edges are sorted by y)

		float minx, maxx, miny, maxy;
		if (p2.x < p1.x) {
			minx = p2.x;
			maxx = p1.x;
		} else {
			minx = p1.x;
			maxx = p2.x;
		}
		if (p2.y < p1.y) {
			miny = p2.y;
			maxy = p1.y;
		} else {
			miny = p1.y;
			maxy = p2.y;
		}

        if(edge.y1 < miny && edge.y2 < miny)
            continue;

        if(edge.y1 > maxy && edge.y2 > maxy)
            continue;

        if(edge.x1 < minx && edge.x2 < minx)
            continue;

        if(edge.x1 > maxx && edge.x2 > maxx)
            continue;

		Vector e1 (edge.x1, edge.y1);
		Vector e2 (edge.x2, edge.y2);

		Vector q1 = p1 - e1;
		Vector q2 = p2 - e1;

		float n1 = q1 | edge.normal;
		if (n1 < 0) continue;

		float n2 = q2 | edge.normal;
		if (n2 > 0) continue;

		Vector linedir (-edge.normal.y, edge.normal.x);

		float len = (e2-e1) | linedir;
		if (len < 0) {
			len = -len;
			linedir.set (edge.normal.y, -edge.normal.x);
		}

		float d1 = q1 | linedir;
		float d2 = q2 | linedir;

		float t = d1 + (d2 - d1) * n1 / (n1 - n2);

		if (t > -tolerance && t < tolerance + len) {
			float d = n1 / (n1 - n2);
			if (dist < -0.5 || d < dist) {
				dist = d;
				depth = -n2;
				coll_edge = &edge;
			}
        }
        
    }

	if (coll_edge) {
        collisions.push_back(LevelCollision(depth,coll_edge->normal));
		return true;
	} else {
		return false;
	}
}


void keep_points_inside_world(Vector* oldpoints, Vector* points, int numpoints)
{
    for(int i=0; i<numpoints; i++) {
        Vector& p1 = oldpoints[i];
        Vector& p2 = points[i];

        vector<Edge>::iterator it;

        float dist = -1; // distance until the collision, used to find the nearest collision only
        Edge *coll_edge = NULL; // nearest colliding edge
        float depth = 0;


        for(it = edges.begin(); it != edges.end(); it++) {
            Edge& edge = *it;
            
            
            // Early giveups (do y first, since edges are sorted by y)
            
            float minx, maxx, miny, maxy;
            if (p2.x < p1.x) {
                minx = p2.x;
                maxx = p1.x;
            } else {
                minx = p1.x;
                maxx = p2.x;
            }
            if (p2.y < p1.y) {
                miny = p2.y;
                maxy = p1.y;
            } else {
                miny = p1.y;
                maxy = p2.y;
            }
            
            if(edge.y1 < miny && edge.y2 < miny)
                continue;
            
            if(edge.y1 > maxy && edge.y2 > maxy)
                continue;
            
            if(edge.x1 < minx && edge.x2 < minx)
                continue;
            
            if(edge.x1 > maxx && edge.x2 > maxx)
                continue;
            
            Vector e1 (edge.x1, edge.y1);
            Vector e2 (edge.x2, edge.y2);

            Vector q1 = p1 - e1;
            Vector q2 = p2 - e1;
            
            float n1 = q1 | edge.normal;
            if (n1 < 0) continue;
            
            float n2 = q2 | edge.normal;
            if (n2 > 0) continue;
            
            Vector linedir (-edge.normal.y, edge.normal.x);
            
            float len = (e2-e1) | linedir;
            if (len < 0) {
                len = -len;
                linedir.set (edge.normal.y, -edge.normal.x);
            }
            
            float d1 = q1 | linedir;
            float d2 = q2 | linedir;
            
            float t = d1 + (d2 - d1) * n1 / (n1 - n2);
            
            p2 = e1 + linedir * t + edge.normal;
        }
        
    }
}
