// lighting routines

#include "lighting.h"
#include "plasma.h"
#include "heart.h"
#include "level.h"
#include "rocket.h"
#include "arrow.h"



// user settings
bool use_lights = true;
bool use_bumpmap = true;



// generated per-frame to hold lighting info
static BITMAP *light_map;
static BITMAP *vector_map;



// precalculated maps holding the falloff curve around a lightsource
static BITMAP *big_light;
static BITMAP *ickle_light;

static BITMAP *big_vectors;
static BITMAP *ickle_vectors;



// normal-map version of the plasma texture
static BITMAP *bumpmap;



// precalculated color transform functions
static COLOR_MAP *greyscale_add_table;
static COLOR_MAP *lightdir_add_table;
static COLOR_MAP *bumpmap_dot_table;
static COLOR_MAP *light_table;



// table for combining two light intensities (precalculated a*b)
static unsigned char mul_table[64][64];



// normalizes a vector
static void normalize(float &x, float &y, float &z)
{
	float d = sqrt(x*x + y*y + z*z);
	
	if (d > 0.00001)
		d = 1.0 / d;
	
	x *= d;
	y *= d;
	z *= d;
}



// packs a light direction vector into an 8 bit value
static unsigned char encode_light_vector(float x, float y, float z, float a, int dx, int dy)
{
	// encode (and dither) the alpha value (banding? nah, 2 bits is enough for anyone :-)
	if (a < 0)
		a = 0;

	int ia2 = (int)(a*3*8);
	int ia = ia2/8;
	
	static unsigned char dither_table[7] = { 16, 68, 146, 170, 109, 187, 239 };
	static unsigned char dither_ytable[8] = { 1, 5, 2, 7, 4, 0, 6, 3 };

	dy = dither_ytable[dy & 7];

	if (ia2 & 7) {
		int dpos = (dx+dy) & 7; 
		ia += ((dither_table[((ia2&7)-1)]) & (1 << dpos)) >> dpos;
	}

	ia = MID(0, ia, 3) << 6;

	// encode the direction
	normalize(x, y, z);

	float ang = atan2(y, x);
	
	if (ang < 0)
		ang += M_PI*2;

	return ((int)(ang * 64.0 / (M_PI*2)) & 0x3F) | ia;
}



// turns an 8 bit value back into a light direction vector
static void decode_light_vector(unsigned char d, float &x, float &y, float &z, float &a)
{
	float ang = (float)(d & 0x3F) * M_PI*2 / 64.0;

	x =  cos(ang);
	y = -sin(ang);
	z =  0;

	a = (float)(d >> 6) / 3.0;
}



// packs a bumpmap direction vector into an 8 bit value
static unsigned char encode_bump_vector(float x, float y, float z)
{
	normalize(x, y, z);

	int ix = (int)(x*7);
	int iy = (int)(y*7);

	ix &= 15;
	iy &= 15;

	return ix | (iy << 4);
}



// turns an 8 bit value back into a bumpmap vector
static void decode_bump_vector(unsigned char d, float &x, float &y, float &z)
{
	int ix = d & 15;
	int iy = d >> 4;

	if (ix & 8)
		ix |= 0xFFFFFFF0;

	if (iy & 8)
		iy |= 0xFFFFFFF0;

	x =  (float)ix / 7.0;
	y = -(float)iy / 7.0;
	z =  0;
}



// creates a texture holding a light falloff curve
static BITMAP *create_light_map(int size)
{
	BITMAP *bmp = create_bitmap(size, size);
	
	for (int y=0; y<size; y++) {
		for (int x=0; x<size; x++) {
			float d = sqrt((x-size/2)*(x-size/2) + (y-size/2)*(y-size/2));

			d /= (float)(size/2);

			d = 1.0 - d;

			bmp->line[y][x] = MID(0, (int)(d*63), 63);
		}
	}
	
	return bmp;
}



// creates a texture holding the direction vectors for a light
static BITMAP *create_light_vectors(int size, bool ambient)
{
	BITMAP *bmp = create_bitmap(size, size);
	
	for (int y=0; y<size; y++) {
		for (int x=0; x<size; x++) {
			float d = 0;
			
			if (ambient) {
				d = sqrt((x-size/2)*(x-size/2) + (y-size/2)*(y-size/2));

				d /= (float)(size/2);

				d = 1.0 - d;
			}

			bmp->line[y][x] = encode_light_vector(x-size/2, y-size/2, 0, d, x, y);
		}
	}
	
	return bmp;
}



// turns a heightfield texture into a bumpmap
static BITMAP *create_bumpmap(BITMAP *src)
{
	BITMAP *bmp = create_bitmap(src->w, src->h);
	
	for (int y=0; y<src->h; y++) {
		for (int x=0; x<src->w; x++) {
			// this pixel height
			float h1 = src->line[y][x];

			// one to the right
			float h2;

			int xx = x;
			
			do {
				xx++;
				xx %= src->w;

				h2 = src->line[y][xx];

			} while ((h2 == h1) && (xx != x));

			// one below
			float h3;
			
			int yy = y;
			
			do {
				yy++;
				yy %= src->h;
				
				h3 = src->line[yy][x];

			} while ((h3 == h1) && (yy != y));

			const float steepness = 0.5;

			// work out the surface right vector
			float right_x = ((xx-x) < 0) ? xx-x+src->w : xx-x;
			float right_y = 0;
			float right_z = (h2 - h1) * steepness;

			// work out the surface down vector
			float down_x = 0;
			float down_y = ((yy-y) < 0) ? yy-y+src->h : yy-y;
			float down_z = (h3 - h1) * steepness;

			// cross them to get the normal
			float normal_x = right_y*down_z - right_z*down_y;
			float normal_y = right_z*down_x - right_x*down_z;
			float normal_z = right_x*down_y - right_y*down_x;

			// store
			bmp->line[y][x] = encode_bump_vector(normal_x, normal_y, normal_z);
		}
	}
	
	return bmp;
}



// lighting color blend function
static void light_blend(const PALETTE pal, int x, int y, RGB *rgb)
{
	rgb->r = MIN((int)pal[y].r * x / 128, 63);
	rgb->g = MIN((int)pal[y].g * x / 128, 63);
	rgb->b = MIN((int)pal[y].b * x / 128, 63);
}



// sets up the lighting system
void init_lighting()
{
	int x, y;

	light_map  = create_bitmap(SCREEN_W, SCREEN_H);
	vector_map = create_bitmap(SCREEN_W, SCREEN_H);

	big_light   = create_light_map(666);
	ickle_light = create_light_map(333);

	big_vectors   = create_light_vectors(666, false);
	ickle_vectors = create_light_vectors(333, true);
	
	bumpmap = create_bumpmap(plasma);
	
	// a+b table
	greyscale_add_table = new COLOR_MAP;

	for (x=0; x<256; x++) {
		for (y=0; y<256; y++)
			greyscale_add_table->data[x][y] = MIN(x+y, 63);
	}

	// light vector addition table
	lightdir_add_table = new COLOR_MAP;

	for (x=0; x<256; x++) {
		for (y=0; y<256; y++) {
			int d1 = x & 0x3F;
			int d2 = y & 0x3F;

			int a1 = x >> 6;
			int a2 = y >> 6;

			int a = MIN(a1+a2, 3);

			lightdir_add_table->data[x][y] = d2 | (a << 6);
		}
	}

	// a|b table
	bumpmap_dot_table = new COLOR_MAP;

	for (x=0; x<256; x++) {
		for (y=0; y<256; y++) {
			float dx, dy, dz, da;
			float bx, by, bz;

			decode_light_vector(x, dx, dy, dz, da);
			decode_bump_vector(y, bx, by, bz);

			float dot = dx*bx + dy*by + dz*bz - da;
			
			bumpmap_dot_table->data[x][y] = MID(0, (int)(-dot*63), 63);
		}
	}

	// sprite lighting table
	light_table = new COLOR_MAP;

	create_color_table(light_table, our_palette, light_blend, NULL);

	// a*b table
	for (x=0; x<64; x++) {
		for (y=0; y<64; y++)
			mul_table[x][y] = MIN(x*y/63+2, 63);
	}
}



// sets up the light maps at the start of each frame
void render_lights()
{
	if (!use_lights)
		return;

	// draw the player light into the light map
	int size = big_light->w;

	int x = (int)(heart->mPos.x + scroll.x);
	int y = (int)(heart->mPos.y + scroll.y);

	blit(big_light, light_map, 0, 0, x-size/2, y-size/2, size, size);
	
	if (use_bumpmap)
		blit(big_vectors, vector_map, 0, 0, x-size/2, y-size/2, size, size);
	
	// clear the area around the player light
	if ((x-size/2) > 0) {
		rectfill(light_map, 0, y-size/2, x-size/2-1, y+size/2-1, 0);

		if (use_bumpmap)
			rectfill(vector_map, 0, y-size/2, x-size/2-1, y+size/2-1, 0);
	}

	if ((x+size/2) < SCREEN_W) {
		rectfill(light_map, x+size/2, y-size/2, SCREEN_W, y+size/2-1, 0);

		if (use_bumpmap)
			rectfill(vector_map, x+size/2, y-size/2, SCREEN_W, y+size/2-1, 0);
	}

	if ((y-size/2) > 0) {
		rectfill(light_map, 0, 0, SCREEN_W-1, y-size/2-1, 0);

		if (use_bumpmap)
			rectfill(vector_map, 0, 0, SCREEN_W-1, y-size/2-1, 0);
	}

	if ((y+size/2) < SCREEN_H) {
		rectfill(light_map, 0, y+size/2, SCREEN_W-1, SCREEN_H-1, 0);

		if (use_bumpmap)
			rectfill(vector_map, 0, y+size/2, SCREEN_W-1, SCREEN_H-1, 0);
	}

	// draw other lights
	int dx = (int)(scroll.x - ickle_light->w/2);
	int dy = (int)(scroll.y - ickle_light->h/2);

	for (ThingList<Rocket>::iterator it = rockets.begin(); it != rockets.end(); it++) {
		color_map = greyscale_add_table;
		draw_trans_sprite(light_map, ickle_light, it->mPos.x+dx, it->mPos.y+dy);

		if (use_bumpmap) {
			color_map = lightdir_add_table;
			draw_trans_sprite(vector_map, ickle_vectors, it->mPos.x+dx, it->mPos.y+dy);
		}
	}

	for (ThingList<Arrow>::iterator it = arrows.begin(); it != arrows.end(); it++) {
		color_map = greyscale_add_table;
		draw_trans_sprite(light_map, ickle_light, it->mPos.x+dx, it->mPos.y+dy);

		if (use_bumpmap) {
			color_map = lightdir_add_table;
			draw_trans_sprite(vector_map, ickle_vectors, it->mPos.x+dx, it->mPos.y+dy);
		}
	}
}



// worker function for drawing a row of the background
void fill_scanline(BITMAP *bmp, int x, int y, int w, int tx, int ty)
{
	unsigned char *dest = bmp->line[y] + x;

	if (use_bumpmap) {
		// full bumpmapped lighting mode
		unsigned char *light = light_map->line[y] + x;
		unsigned char *dir = vector_map->line[y] + x;
		unsigned char *bump = bumpmap->line[ty & 255];
		
		do {
			*(dest++) = mul_table[*(light++)][bumpmap_dot_table->data[*(dir++)][bump[tx++ & 255]]];
		} while (w--);
	}
	else if (use_lights) {
		// simple lighting mode
		unsigned char *light = light_map->line[y] + x;
		unsigned char *tex = plasma->line[ty & 255];
		
		do {
			*(dest++) = mul_table[*(light++)][tex[tx++ & 255]];
		} while (w--);
	}
	else {
		// easy, just draw the plasma texture
		unsigned char *tex = plasma->line[ty & 255];
		
		do {
			*(dest++) = tex[tx++ & 255];
		} while (w--);
	}
}



// looks up a point in the light map
static int light_sample(int x, int y)
{
	if (x < 0)
		x = 0;
	else if (x >= SCREEN_W)
		x = SCREEN_W-1;

	if (y < 0)
		y = 0;
	else if (y >= SCREEN_H)
		y = SCREEN_H-1;

	return MID(0, light_map->line[y][x]*255/63, 255);
}



// looks up a point in the light map, plus separate ambiency
static int light_sample(int x, int y, float &a)
{
	if (x < 0)
		x = 0;
	else if (x >= SCREEN_W)
		x = SCREEN_W-1;

	if (y < 0)
		y = 0;
	else if (y >= SCREEN_H)
		y = SCREEN_H-1;

	a = 0;

	int c = 0;

	for (int dx=-1; dx<=1; dx++) {
		for (int dy=-1; dy<=1; dy++) {
			int ddx = MID(0, x+dx, SCREEN_W-1);
			int ddy = MID(0, y+dy, SCREEN_H-1);

			a += (float)(vector_map->line[ddy][ddx] >> 6) / 3.0;
			
			c++;
		}
	}

	a /= c;

	return MID(0, light_map->line[y][x]*255/63, 255);
}



// lit sprite renderer
void draw_lit_dude(BITMAP *bmp, BITMAP *sprite, int x, int y)
{
	if (use_bumpmap) {
		// us is bumpmapped
		color_map = light_table;

		Vector pos(x+sprite->w/2, y+sprite->h/2);

		Vector dir = pos - heart->mPos - scroll;

		float dist = dir.magnitude();

		dir.normalize();

		float a;
		
		float l = light_sample(pos.x, pos.y, a);

		float c[4];
		
		for (int i=0; i<4; i++) {
			static Vector normals[4] =
			{
				~Vector( 1,  1),
				~Vector(-1,  1),
				~Vector(-1, -1),
				~Vector( 1, -1),
			};
			
			float dot = dir | normals[i];
			
			if (dot < 0)
				dot = 0;
			
			c[i] = (dot + a) * l;
			
			c[i] = MID(0, c[i], 255);
		}

		draw_gouraud_sprite(bmp, sprite, x, y, (int)c[0], (int)c[1], (int)c[2], (int)c[3]);
	}
	else if (use_lights) {
		// pooh, us is just lit regular like
		color_map = light_table;

		draw_lit_sprite(bmp, sprite, x, y, light_sample(x+sprite->w/2, y+sprite->w/2));
	}
	else {
		// simple regular mode
		draw_sprite(bmp, sprite, x, y);
	}
}


